<?php

use YooKassa\Model\CancellationDetailsPartyCode;
use YooKassa\Model\CancellationDetailsReasonCode;
use YooKassa\Model\Confirmation\AbstractConfirmation;
use YooKassa\Model\ConfirmationType;
use YooKassa\Model\Notification\AbstractNotification;
use YooKassa\Model\Notification\NotificationFactory;
use YooKassa\Model\Notification\NotificationRefundSucceeded;
use YooKassa\Model\NotificationEventType;
use YooKassa\Model\PaymentMethodType;
use YooKassa\Model\PaymentStatus;

/**
 * Class ControllerPaymentYoomoney
 *
 * @property-read Language $language
 * @property-read Currency $currency
 *
 * @property ModelCheckoutOrder $model_checkout_order
 * @property Cart $cart
 */
class ControllerPaymentYoomoney extends Controller
{
    /**
     * @var ModelPaymentYoomoney Модель работы с платежами
     */
    private $_model;

    /**
     * @var array Массив с информацией о текущем заказе
     */
    private $_orderInfo;

    /**
     * Экшен отображения страницы выбора способа оплаты
     */
    protected function index()
    {
        $this->payment($this->getOrderInfo());
    }

    /**
     * Экшен создания платежа на стороне ЮKassa
     *
     * После создания осуществляет редирект на страницу оплаты на стороне кассы. Доступен только для способа оплаты
     * через кассу, для платежей в кошелёк и с помощью платёжки просто редиректит на страницу корзины.
     */
    public function create()
    {
        $orderInfo = $this->getOrderInfo('order_id', false);
        if ($orderInfo === null) {
            $this->jsonError('Корзина пуста');
        }
        $this->language->load('payment/yoomoney');
        $this->getModel()->log('info', 'Создание платежа для заказа №'.$orderInfo['order_id']);
        /** @var YooMoneyPaymentKassa $paymentMethod */
        $paymentMethod = $this->getModel()->getPaymentMethod($this->config->get('yoomoney_mode'));
        if (!$paymentMethod->isModeKassa()) {
            $this->jsonError('Ошибка настройки модуля');
        }

        $successUrl = str_replace(
            array('&amp;'),
            array('&'),
            $this->url->link('payment/yoomoney/confirm', 'order_id='.$orderInfo['order_id'], 'SSL')
        );


        $payment = $this->getModel()->createPayment($paymentMethod, $orderInfo);
        if ($payment === null) {
            $this->jsonError('Платеж не прошел. Попробуйте еще или выберите другой способ оплаты');
        };

        $result = array(
            'success'  => true,
            'redirect' => $successUrl,
        );

        /** @var AbstractConfirmation $confirmation */
        $confirmation = $payment->getConfirmation();
        if ($confirmation === null) {
            $this->getModel()->log('warning', 'Confirmation in created payment equals null');
        } elseif ($confirmation->getType() === ConfirmationType::REDIRECT) {
            $result['redirect'] = $confirmation->getConfirmationUrl();
        }

        if ($paymentMethod->getCreateOrderBeforeRedirect()) {
            $this->getModel()->confirmOrder($paymentMethod, $orderInfo['order_id']);
        }
        if ($paymentMethod->getClearCartBeforeRedirect()) {
            $this->cart->clear();
        }

        echo json_encode($result);
        exit();
    }

    /**
     * Экшен подтверждения платежа, вызывается при возврате пользователя из кассы
     *
     * Очищает корзину, устанавливает нужный статус заказа, если нужно, осуществляет подтверждение платежа на стороне
     * кассы. Если платёж в статусе кансэллед, то редиректит на страницу ошибки.
     */
    public function confirm()
    {
        $paymentMethod = $this->getModel()->getPaymentMethod($this->config->get('yoomoney_mode'));
        if ($paymentMethod instanceof YooMoneyPaymentKassa) {

            if (!isset($_GET['order_id'])) {
                $this->errorRedirect('Order id not specified in return link');
            }

            $this->language->load('payment/yoomoney');
            $orderId = (int)$_GET['order_id'];
            if ($orderId <= 0) {
                $this->errorRedirect('Invalid order id in return link: '.json_encode($_GET['order_id']));
            }

            $this->getModel()->log('info', 'Возврат пользователя из кассы для заказа №'.$orderId);
            $payment = $this->getModel()->getPaymentByOrderId($paymentMethod, $orderId);
            if ($payment === null) {
                $this->redirect($this->url->link('checkout/checkout', '', 'SSL'));
            } elseif ($payment->getStatus() === \YooKassa\Model\PaymentStatus::CANCELED) {
                $pageId      = $this->config->get('yoomoney_kassa_page_failure');
                $redirectUrl = (empty($pageId) || $pageId < 0)
                    ? $this->url->link('checkout/checkout', '', 'SSL')
                    : $this->url->link('information/information', 'information_id='.$pageId, 'SSL');
                $this->redirect($redirectUrl);
            } elseif (!$payment->getPaid()) {
                $this->redirect($this->url->link('checkout/checkout', '', 'SSL'));
            }

            $pageId = $this->config->get('yoomoney_kassa_page_success');

            if (isset($this->session->data['order_id']) && $orderId === $this->session->data['order_id']) {
                $this->cart->clear();
            }
            $redirectUrl = (empty($pageId) || $pageId < 0)
                ? $this->url->link('checkout/success', 'order_id='.$orderId, 'SSL')
                : $this->url->link('information/information', 'information_id='.$pageId, 'SSL');

            $this->redirect($redirectUrl);

        }

        $this->jsonError('Unknown payment type');
    }

    /**
     * Экшен подтверждения платежа, дергается API после холдирования
     */
    public function capture()
    {
        $this->language->load('payment/yoomoney');
        $data = file_get_contents('php://input');
        if (empty($data)) {
            $log = 'Empty body in capture notification, get: '.json_encode($_GET).', post: '.json_encode($_POST);
            $this->getModel()->log('error', $log);
            header('HTTP/1.1 400 Empty body in notification request');
            exit();
        }
        $json = @json_decode($data, true);
        if (empty($json)) {
            if (json_last_error() === JSON_ERROR_NONE) {
                $this->getModel()->log('error', 'Empty object in body in capture notification');
            } else {
                $this->getModel()->log('error',
                    'Invalid body in capture notification '.json_last_error_msg().' '.$data);
            }
            header('HTTP/1.1 400 Failed to parse body');
            exit();
        }

        $this->getModel()->log('info', 'Notification: '.$data);

        /** @var YooMoneyPaymentKassa $paymentMethod */
        $paymentMethod = $this->getModel()->getPaymentMethod($this->config->get('yoomoney_mode'));
        if (!$paymentMethod->isModeKassa()) {
            $this->getModel()->log('warning', 'Invalid body in capture notification '.json_last_error_msg().' '.$data);
            header('HTTP/1.1 405 Invalid order payment method');
            exit();
        }

        try {
            $factory = new NotificationFactory();
            $notification = $factory->factory($json);
        } catch (\Exception $e) {
            $this->getModel()->log('error', 'Invalid notification object - '.$e->getMessage());
            header('HTTP/1.1 400 Invalid object in body');
            exit();
        }

        $metadata = $notification->getObject()->getMetadata();

        if (
            isset($metadata['cms_name'])
            && !in_array($metadata['cms_name'], array(ModelPaymentYoomoney::CMS_NAME, ModelPaymentYoomoney::CMS_NAME_OLD))
        ) {
            $this->getModel()->log('info', 'This notification not for opencart1.5. This notification for: ' . $metadata['cms_name']);
            header('HTTP/1.1 400 Invalid object in body');
            exit();
        }

        $paymentId = $notification instanceof NotificationRefundSucceeded
                   ? $notification->getObject()->getPaymentId()
                   : $notification->getObject()->getId();

        $orderId = $this->getModel()->getOrderIdByPayment($paymentId);
        if ($orderId <= 0) {
            $this->getModel()->log('error', 'Order not exists for payment ' . $paymentId);
            exit();
        }

        if ($notification->getEvent() === NotificationEventType::REFUND_SUCCEEDED) {
            $this->getModel()->log('info', 'Refund success for order #'.$orderId);
            exit();
        }

        if ($notification->getEvent() === NotificationEventType::PAYMENT_CANCELED) {
            $notifyPaymentMethod = $notification->getObject()->getPaymentMethod();
            $canceledStatusId = $paymentMethod->getCancelOrderStatusId();
            $cancellationDetails = $notification->getObject()->getCancellationDetails();
            $changeStatus = $cancellationDetails
                && $cancellationDetails->getParty() === CancellationDetailsPartyCode::MERCHANT
                && $cancellationDetails->getReason() === CancellationDetailsReasonCode::CANCELED_BY_MERCHANT;
            if (
                $notifyPaymentMethod
                && $notifyPaymentMethod->getType() === PaymentMethodType::SBER_LOAN
                && $notifyPaymentMethod->getDiscountAmount()
            ) {
                $this->load->model('checkout/order');
                $orderInfo = $this->model_checkout_order->getOrder($orderId);
                $statusId = $changeStatus ? $canceledStatusId : $orderInfo['order_status_id'];
                try {
                    $this->handleSberLoan($notification, $orderInfo, $statusId);
                } catch (Exception $e) {
                    $this->getModel()->log(
                        'error',
                        'Sber loan. Discount apply error',
                        ['exception' => $e]
                    );
                }
            }

            if ($changeStatus) {
                $this->getModel()->updateOrderStatus($orderId, array('order_status_id' => $canceledStatusId));
                $this->_model->updateOrderHistory(
                    $orderId,
                    $canceledStatusId,
                    $this->language->get('captures_cancel_success')
                );
                $this->getModel()->updatePaymentStatus($paymentId, PaymentStatus::CANCELED);
            }
            $this->getModel()->log('info', 'Payment for order #'.$orderId.' cancelled');
            exit();
        }

        $this->load->model('checkout/order');
        $orderInfo = $this->model_checkout_order->getOrder($orderId);
        if (empty($orderInfo)) {
            $this->getModel()->log('warning', 'Empty order #'.$orderId.' in notification');
            exit();
        } elseif ($orderInfo['order_status_id'] <= 0) {
            $this->getModel()->confirmOrder($paymentMethod, $orderId);
        }

        $this->getModel()->log('info',
            'Пришла нотификация для платежа '.$notification->getObject()->getId().' для заказа №'.$orderId);

        $client = $this->getModel()->getClient($paymentMethod);

        try {
            $payment = $client->getPaymentInfo($notification->getObject()->getId());
        } catch (Exception $e) {
            $this->getModel()->log('error',
                'Payment '.$notification->getObject()->getId().' not fetched from API in capture method. Error: ' . $e->getMessage());
            exit();
        }

        if ($notification->getEvent() === NotificationEventType::PAYMENT_WAITING_FOR_CAPTURE
            && $payment->getStatus() === PaymentStatus::WAITING_FOR_CAPTURE
        ) {
            $comment = sprintf($this->language->get('captures_new_hold_payment'),
                $payment->getExpiresAt()->format('d.m.Y H:i'));
            $this->getModel()->log('info', $comment);
            $this->model_checkout_order->update($orderId, $paymentMethod->getHoldOrderStatusId(), $comment);

            exit();
        }

        if (
            $notification->getEvent() === NotificationEventType::PAYMENT_SUCCEEDED
            && $payment->getStatus() === PaymentStatus::SUCCEEDED
        ) {
            $notifyPaymentMethod = $notification->getObject()->getPaymentMethod();
            if (
                $notifyPaymentMethod->getType() === PaymentMethodType::SBER_LOAN
                && $notifyPaymentMethod->getDiscountAmount()
            ) {
                try {
                    $this->handleSberLoan($notification, $orderInfo, $paymentMethod->getOrderStatusId());
                } catch (Exception $e) {
                    $this->getModel()->log(
                        'error',
                        'Sber loan. Discount apply error',
                        ['exception' => $e]
                    );
                    exit();
                }
            }
            $this->getModel()->hookOrderStatusChange($orderId, $paymentMethod->getOrderStatusId());
            $this->getModel()->confirmOrderPayment($orderId, $payment, $paymentMethod->getOrderStatusId());
            $this->getModel()->log('info', 'Платёж для заказа №'.$orderId.' подтверждён');
            exit();
        }
        exit();
    }

    /**
     * @param AbstractNotification $notification
     * @param array $orderInfo
     * @param int $statusId
     * @return void
     * @throws Exception
     */
    private function handleSberLoan($notification, $orderInfo, $statusId)
    {
        $NotifyPaymentMethod = $notification->getObject()->getPaymentMethod();
        $paymentId = $notification->getObject()->getId();

        $discountAmount = $NotifyPaymentMethod->getDiscountAmount()->value;
        $this->getModel()->log(
            'info',
            'Sber loan. Discount amount: ' . $discountAmount
        );
        $this->addDiscountToOrder($orderInfo, $discountAmount);

        $paymentAmount = $notification->getObject()->getAmount()->getValue();
        $this->getModel()->updatePaymentAmount($paymentId, $paymentAmount);

        $this->_model->updateOrderHistory(
            $orderInfo['order_id'],
            $statusId,
            $this->language->get('text_history_sber_loan_discount') . $discountAmount
        );
    }

    /**
     * @param $orderInfo
     * @param $discountAmount
     * @return void
     * @throws Exception
     */
    private function addDiscountToOrder($orderInfo, $discountAmount)
    {
        $discount = (float)str_replace(' ', '', $discountAmount);
        $orderTotal =  $orderInfo['total'];

        /** Высчитываем процент, по которому будем вычислять скидку для всех позиций в заказе */
        $percentDiscount = round(100 / ($orderTotal / (int)$discount), 1);

        $orderId = $orderInfo['order_id'];

        $this->load->model('account/order');
        $this->load->model('account/customer');
        $products = $this->model_account_order->getOrderProducts($orderId);
        $totals = $this->model_account_order->getOrderTotals($orderId);

        $isTaxInOrder = false;
        $isDeliveryInOrder = false;
        foreach($totals as $total) {
            if ($total['code'] === 'tax') {
                $isTaxInOrder = true;
            }

            if ($total['code'] === 'shipping') {
                $isDeliveryInOrder = true;
            }
        }

        $taxAndShippingInOrder = $isTaxInOrder && $isDeliveryInOrder;

        $orderTotalCalculated = $this->calculateTotalFromPercent($orderInfo['total'], $percentDiscount);
        $this->db->query("UPDATE `".DB_PREFIX."order` SET `total` = " . $orderTotalCalculated . " WHERE `order_id` = "
            . (int)$orderInfo['order_id']);

        foreach ($totals as $total) {
            if ($taxAndShippingInOrder && $total['code'] === 'total') {
                $this->updateTextAndTotalWithDiscount($total, $orderInfo, $percentDiscount);
            }

            if (!$taxAndShippingInOrder && $total['value']) {
                $this->updateTextAndTotalWithDiscount($total, $orderInfo, $percentDiscount);
            }
        }

        foreach ($products as $product) {
            if (!$taxAndShippingInOrder) {
                $total = $this->calculateTotalFromPercent($product['total'], $percentDiscount);
                $tax = $this->calculateTotalFromPercent($product['tax'], $percentDiscount);
                $this->db->query("UPDATE `".DB_PREFIX."order_product` SET `total` = "
                    . $total . ", `tax` = " . $tax . " WHERE `order_product_id` = "
                    . (int)$product['order_product_id']);
            }
        }
    }

    /**
     * @param $total
     * @param $orderInfo
     * @param $percentDiscount
     * @return void
     */
    private function updateTextAndTotalWithDiscount($total, $orderInfo, $percentDiscount)
    {
        $value = $this->calculateTotalFromPercent($total['value'], $percentDiscount);
        $text = $this->getAmountWithCurrency($orderInfo['currency_code'], $value);
        $this->db->query("UPDATE `".DB_PREFIX."order_total` SET `value` = "
            . $value . ", `text` = '" . $text . "' WHERE `order_total_id` = "
            . (int)$total['order_total_id']);
    }

    /**
     * Вычисление процента из числа
     *
     * @param float $total Общая сумма
     * @param float $percent Процент, который хотим вычесть из числа
     *
     * @return float
     */
    private function calculateTotalFromPercent($total, $percent)
    {
        return $total - ($percent * ($total / 100));
    }

    /**
     * @param $currencyCode
     * @param $value
     * @return string|void
     */
    private function getAmountWithCurrency($currencyCode, $amount)
    {
        $this->load->model('localisation/currency');

        $currencyData = $this->model_localisation_currency->getCurrencyByCode($currencyCode);

        if (!empty($currencyData['symbol_left'])) {
            return $currencyData['symbol_left']
                . number_format($amount, 2, '.', '');
        }

        if (!empty($currencyData['symbol_right'])) {
            return number_format($amount, 2, '.', '')
                . $currencyData['symbol_right'];
        }
    }

    /**
     * Метод отображения способов оплаты пользователю
     *
     * @param $orderInfo
     * @param bool $child
     */
    private function payment($orderInfo, $child = false)
    {
        $this->language->load('payment/yoomoney');
        $paymentMethod = $this->getModel()->getPaymentMethod($this->config->get('yoomoney_mode'));

        if (isset($orderInfo['email'])) {
            $this->data['email'] = $orderInfo['email'];
        }
        if (isset($orderInfo['telephone'])) {
            $this->data['phone'] = $orderInfo['telephone'];
        }
        if ($this->currency->has('RUB')) {
            $this->data['sum'] = sprintf('%.2f', $this->currency->format($orderInfo['total'], 'RUB', '', false));
        } else {
            $this->data['sum'] = sprintf('%.2f', $this->getModel()->convertFromCbrf($orderInfo, 'RUB'));
        }
        $this->data['cmsname']        = ($child) ? 'opencart-extracall' : 'opencart';
        $this->data['button_confirm'] = $this->language->get('button_confirm');
        $this->data['order_id']       = $orderInfo['order_id'];
        $this->data['paymentMethod']  = $paymentMethod;
        $this->data['lang']           = $this->language;
        $this->assignKassa($paymentMethod);

        if (file_exists(DIR_TEMPLATE.$this->config->get('config_template').'/template/payment/yoomoney.tpl')) {
            $this->template = $this->config->get('config_template').'/template/payment/yoomoney.tpl';
        } else {
            $this->template = 'default/template/payment/yoomoney.tpl';
        }
        if ($child) {
            $this->children = array(
                'common/column_left',
                'common/column_right',
                'common/footer',
                'common/header',
            );
        }

        $this->response->addHeader('Content-Type: text/html; charset=utf-8');
        $this->response->setOutput($this->render());
    }

    public function jsonError($message)
    {
        $this->getModel()->log('warning', 'Error in json: '.$message);
        echo json_encode(array(
            'success' => false,
            'error'   => $message,
        ));
        exit();
    }

    public function assignKassa(YooMoneyPaymentKassa $paymentMethod)
    {
        $this->data['tpl'] = 'kassa';

        $this->data['allow_methods']  = array();
        $this->data['default_method'] = $this->config->get('yoomoney_kassa_payment_default');
        $this->data['validate_url'] = $this->url->link('payment/yoomoney/create', '', 'SSL');

        if (isset($this->request->server['HTTPS']) && (($this->request->server['HTTPS'] == 'on') || ($this->request->server['HTTPS'] == '1'))) {
            $this->data['imageurl'] = $this->config->get('config_ssl').'image/';
        } else {
            $this->data['imageurl'] = $this->config->get('config_url').'image/';
        }

        $title = $this->config->get('yoomoney_kassa_payment_method_name');
        if (empty($title)) {
            $title = $this->language->get('kassa_page_title_default');
        }
        $this->data['method_label'] = $title;
    }

    public function repay()
    {
        if (!$this->customer->isLogged()) {
            $this->session->data['redirect'] = $this->url->link('payment/yoomoney/repay',
                'order_id='.$this->request->get['order_id'], 'SSL');
            $this->redirect($this->url->link('account/login', '', 'SSL'));
        }
        $this->load->model('account/order');
        $order_info = $this->model_account_order->getOrder((int)$this->request->get['order_id']);
        if ($order_info) {
            $this->payment($order_info, true);
        } else {
            $this->redirect($this->url->link('account/order/info', 'order_id='.$this->request->get['order_id'], 'SSL'));
        }
    }

    public function success()
    {
        if (isset($_GET['order_id'])) {
            $this->session->data['tmp_order_id'] = (int)$_GET['order_id'];
            $orderInfo                           = $this->getOrderInfo('tmp_order_id');
            $this->data['order']                 = $orderInfo;
        }
        $this->renderPage('success', true);
    }

    public function failure()
    {
        if (isset($_GET['order_id'])) {
            $this->session->data['tmp_order_id'] = (int)$_GET['order_id'];
            $orderInfo                           = $this->getOrderInfo('tmp_order_id');
            $this->data['order']                 = $orderInfo;
        }
        $this->renderPage('failure', true);
    }

    public function renderPage($template, $child = false)
    {
        $templatePath = '/template/payment/yoomoney/'.$template.'.tpl';
        if (file_exists(DIR_TEMPLATE.$this->config->get('config_template').$templatePath)) {
            $this->template = $this->config->get('config_template').$templatePath;
        } else {
            $this->template = 'default'.$templatePath;
        }
        if ($child) {
            $this->children = array(
                'common/column_left',
                'common/column_right',
                'common/footer',
                'common/header',
            );
        }
        $this->response->addHeader('Content-Type: text/html; charset=utf-8');
        $this->response->setOutput($this->render());
    }

    private function updateOrderStatus($orderId, $status, $text)
    {
        $this->model_checkout_order->confirm($orderId, $status, $text);

        $this->cart->clear();
        if (isset($this->session->data['order_id'])) {
            unset($this->session->data['shipping_method']);
            unset($this->session->data['shipping_methods']);
            unset($this->session->data['payment_method']);
            unset($this->session->data['payment_methods']);
            unset($this->session->data['guest']);
            unset($this->session->data['comment']);
            unset($this->session->data['order_id']);
            unset($this->session->data['coupon']);
            unset($this->session->data['reward']);
            unset($this->session->data['voucher']);
            unset($this->session->data['vouchers']);
            unset($this->session->data['totals']);
        }
    }

    private function parsePlaceholders($template, $order)
    {
        $replace = array();
        foreach ($order as $key => $value) {
            if (is_scalar($value)) {
                $replace['%'.$key.'%'] = $value;
            }
        }

        return strtr($template, $replace);
    }

    /**
     * Возвращает модель работы с платежами, если модель ещё не инстацирована, создаёт её
     * @return ModelPaymentYoomoney Модель работы с платежами
     */
    public function getModel()
    {
        if ($this->_model === null) {
            $this->load->model('payment/yoomoney');
            $this->_model = $this->model_payment_yoomoney;
        }

        return $this->_model;
    }

    /**
     * Возвращает информаицю о текущем заказе в корзине, если заказа нет, редиректит на страницу корзины
     *
     * @param string $sessionKey Ключ в сессии, по которому лежит айди заказа
     * @param bool $redirectOnError Требуется ли перенаправить пользователя, если произошла ошибка
     *
     * @return array|null Массив с информацией о платеже или null если произошла ошибка и флаг редиректа равен false
     */
    public function getOrderInfo($sessionKey = 'order_id', $redirectOnError = true)
    {
        if ($this->_orderInfo === null) {
            if (!isset($this->session->data[$sessionKey])) {
                if ($redirectOnError) {
                    $this->errorRedirect('Order id ('.$sessionKey.') not exists in session');
                } else {
                    return null;
                }
            }
            $this->load->model('checkout/order');
            $this->_orderInfo = $this->model_checkout_order->getOrder($this->session->data[$sessionKey]);
            if (empty($this->_orderInfo)) {
                if ($redirectOnError) {
                    $this->errorRedirect('Order#'.$this->session->data[$sessionKey].' not exists in database');
                } else {
                    return null;
                }
            }
        }

        return $this->_orderInfo;
    }

    /**
     * Осуществляет редирект на страницу
     *
     * @param string $message Почему пользователя редиректит
     * @param string $redirectLink Ссылка на страницу редиректа
     */
    public function errorRedirect($message, $redirectLink = 'checkout/cart')
    {
        $this->getModel()->log('warning', 'Redirect user: '.$message);
        $this->redirect($this->url->link($redirectLink));
    }
}
